# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest

from qiime2.plugin.testing import TestPluginBase
from qiime2 import Artifact


class AlignToTreeMafftRaxmlPipelineTest(TestPluginBase):
    package = 'q2_phylogeny.tests'

    def setUp(self):
        super().setUp()
        self.align_to_tree_mafft_raxml = self.plugin.pipelines[
                                    'align_to_tree_mafft_raxml']

        input_sequences_fp = self.get_data_path('dna-sequences-1.fasta')
        self.input_sequences = Artifact.import_data('FeatureData[Sequence]',
                                                    input_sequences_fp)

    def test_outputs(self):
        result = self.align_to_tree_mafft_raxml(self.input_sequences)
        self.assertEqual(4, len(result))
        aligned_seq, masked_seq, unrooted_tree, rooted_tree = result
        self.assertEqual('FeatureData[AlignedSequence]', str(aligned_seq.type))
        self.assertEqual('FeatureData[AlignedSequence]', str(masked_seq.type)),
        self.assertEqual('Phylogeny[Unrooted]', str(unrooted_tree.type)),
        self.assertEqual('Phylogeny[Rooted]', str(rooted_tree.type))


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import skbio


def midpoint_root(tree: skbio.TreeNode) -> skbio.TreeNode:
    return tree.root_at_midpoint()


def robinson_foulds(trees: skbio.TreeNode, labels: str = None,
                    missing_tips: str = 'error') -> skbio.DistanceMatrix:
    if labels is None:
        labels = ['tree_%d' % d for d in range(1, len(trees) + 1)]
    elif len(trees) != len(labels):
        raise ValueError("The number of trees and labels must match.")

    tips = [{t.name for t in tree.tips()} for tree in trees]
    shared_tips = set.intersection(*tips)
    if not shared_tips:
        raise ValueError("No tip names are shared between these trees.")
    if missing_tips == 'intersect-all':
        trees = [t.shear(shared_tips) for t in trees]
    elif missing_tips == 'error':
        all_tips = set.union(*tips)
        if shared_tips != all_tips:
            SKIP = 10
            miss = list(all_tips - shared_tips)
            missing_repr = ", ".join(map(repr, miss[:SKIP]))
            if len(miss) > SKIP:
                missing_repr += ", ...<%d ommitted>" % (len(miss) - SKIP)

            raise ValueError("Not all tips are shared between trees: "
                             + missing_repr)
    else:
        raise ValueError("Unknown argument for missing_tips=%r"
                         % missing_tips)

    return skbio.DistanceMatrix.from_iterable(
        trees, metric=skbio.TreeNode.compare_rfd,
        keys=labels, validate=False)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------


def align_to_tree_mafft_raxml(ctx, sequences, n_threads=1,
                              mask_max_gap_frequency=1.0,
                              mask_min_conservation=0.40,
                              parttree=False,
                              substitution_model='GTRGAMMA',
                              seed=None, raxml_version='Standard'):
    mafft = ctx.get_action('alignment', 'mafft')
    mask = ctx.get_action('alignment', 'mask')
    raxml = ctx.get_action('phylogeny', 'raxml')
    midpoint_root = ctx.get_action('phylogeny', 'midpoint_root')

    aligned_seq, = mafft(sequences=sequences, n_threads=n_threads,
                         parttree=parttree)
    masked_seq, = mask(alignment=aligned_seq,
                       max_gap_frequency=mask_max_gap_frequency,
                       min_conservation=mask_min_conservation)
    unrooted_tree, = raxml(alignment=masked_seq, n_threads=n_threads,
                           substitution_model=substitution_model,
                           seed=seed, raxml_version=raxml_version)
    rooted_tree, = midpoint_root(tree=unrooted_tree)

    return (aligned_seq, masked_seq, unrooted_tree, rooted_tree)



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import subprocess

from q2_types.feature_data import AlignedDNAFASTAFormat
from q2_types.tree import NewickFormat


def run_command(cmd, output_fp, verbose=True, env=None):
    if verbose:
        print("Running external command line application. This may print "
              "messages to stdout and/or stderr.")
        print("The command being run is below. This command cannot "
              "be manually re-run as it will depend on temporary files that "
              "no longer exist.")
        print("\nCommand:", end=' ')
        print(" ".join(cmd), end='\n\n')

    with open(output_fp, 'w') as output_f:
        subprocess.run(cmd, stdout=output_f, check=True, env=env)


def fasttree(alignment: AlignedDNAFASTAFormat,
             n_threads: int = 1) -> NewickFormat:
    result = NewickFormat()
    aligned_fp = str(alignment)
    tree_fp = str(result)

    env = None
    if n_threads == 1:
        cmd = ['FastTree']
    else:
        env = os.environ.copy()
        if n_threads == 0:
            env.pop('OMP_NUM_THREADS', 0)
        else:
            env.update({'OMP_NUM_THREADS': str(n_threads)})
        cmd = ['FastTreeMP']

    cmd.extend(['-quote', '-nt', aligned_fp])
    run_command(cmd, tree_fp, env=env)
    return result



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from itertools import zip_longest
import math
import os
import tempfile
import unittest

import skbio
from qiime2.plugin.testing import TestPluginBase
from qiime2.util import redirected_stdio
from q2_types.feature_data import AlignedDNAFASTAFormat

from q2_phylogeny import iqtree, iqtree_ultrafast_bootstrap
from q2_phylogeny._raxml import run_command
from q2_phylogeny._iqtree import (_build_iqtree_command,
                                  _build_iqtree_ufbs_command)


class IqtreeTests(TestPluginBase):

    package = 'q2_phylogeny.tests'

    def test_iqtree(self):
        # Test that output tree is made.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = iqtree(input_sequences)
        obs_tree = skbio.TreeNode.read(str(obs))
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_iqtree_safe_allnni(self):
        # Same as `test_iqtree` but testing the `-safe` and `-allnni `flags
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = iqtree(input_sequences, safe='True', allnni='True')
        obs_tree = skbio.TreeNode.read(str(obs))
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_iqtree_underscore_ids(self):
        # Test that output tree is made with underscores in tip IDs.
        # Some programs and python wrappers may strip underscores.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-4.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = iqtree(input_sequences)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA_001510755_1', 'GCA_001045515_1',
                              'GCA_000454205_1', 'GCA_000473545_1',
                              'GCA_000196255_1', 'GCA_002142615_1',
                              'GCA_000686145_1', 'GCA_001950115_1',
                              'GCA_001971985_1', 'GCA_900007555_1']))

    def test_iqtree_n_cores(self):
        # Test that an output tree is made when invoking multiple cores.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = iqtree(input_sequences, n_cores=2)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)

        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]

        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_iqtree_n_cores_auto(self):
        # Test that an output tree is made when invoking automatic threads.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = iqtree(input_sequences, n_cores=0)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)

        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]

        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_iqtree_n_cores_auto_max(self):
        # Test that an output tree is made when invoking automatic threads.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = iqtree(input_sequences, n_cores=0, n_cores_max=2)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)

        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]

        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_iqtree_with_seed(self):
        # Test tip-to-tip dists are identical to manually run IQ-TREE output.
        # This test is comparing an ordered series of tip-to-tip distances
        # to a tree output from a manual run of the default command:
        #     iqtree -seed 1723 -m HKY -s aligned-dna-sequences-3.fasta
        #            -nt 1 -pre q2iqtree
        # NOTE: I cleanly rounded the tip-to-tip dists (i.e. `%.4f`) as
        # IQ-TREE may return slightly different rounding errors on different
        # systems.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = iqtree(input_sequences, seed=1723,
                         substitution_model='HKY')
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)
        obs_tl = list(obs_tree.tip_tip_distances().to_series())
        obs_series = set(['%.3f' % e for e in obs_tl])

        exp_tree = skbio.TreeNode.read(self.get_data_path('test4.tre'))
        exp_tl = list(exp_tree.tip_tip_distances().to_series())
        exp_series = set(['%.3f' % e for e in exp_tl])

        self.assertEqual(obs_series, exp_series)

    def test_iqtree_model_choice(self):
        # Tip to tip dists should NOT be identical under different models.
        # Default is MFP (auto select substitution model). We'll compare ouput
        # of the GTR+G and HKY models.
        # This test is comparing an ordered series of tip-to-tip distances.
        # Take note, that for this comparison to work, all must have the same
        # seed value set.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        # default GTR+G
        with redirected_stdio(stderr=os.devnull):
            gtrg = iqtree(input_sequences, seed=1723,
                          substitution_model='GTR+G')
            gtrg_tree = skbio.TreeNode.read(
                        str(gtrg), convert_underscores=False)
            gtrg_td = set(gtrg_tree.tip_tip_distances().to_series())

        # set HKY
        with redirected_stdio(stderr=os.devnull):
            hky = iqtree(input_sequences, seed=1723,
                         substitution_model='HKY')
            hky_tree = skbio.TreeNode.read(
                         str(hky), convert_underscores=False)
            hky_td = set(hky_tree.tip_tip_distances().to_series())

        # test pairs are not equivalent
        self.assertNotEqual(gtrg_td, hky_td)

    def test_build_iqtree_command(self):
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with tempfile.TemporaryDirectory() as temp_dir:
            run_prefix = os.path.join(temp_dir, 'q2iqtree')
            with redirected_stdio(stderr=os.devnull):
                obs = _build_iqtree_command(input_sequences,
                                            seed=1723,
                                            n_cores=0,
                                            n_cores_max=4,
                                            n_runs=2,
                                            substitution_model='MFP',
                                            run_prefix=run_prefix,
                                            dtype='DNA',
                                            safe='True',
                                            fast='True',
                                            alrt=1000,
                                            abayes=True,
                                            lbp=1000,
                                            n_init_pars_trees=200,
                                            n_top_init_trees=30,
                                            n_best_retain_trees=10,
                                            n_iter=80,
                                            stop_iter=300,
                                            perturb_nni_strength=0.55,
                                            spr_radius=8,
                                            allnni='True')
        self.assertTrue('DNA' in obs[2])
        self.assertTrue('2' in obs[4])
        self.assertTrue(str(input_sequences) in str(obs[6]))
        self.assertTrue('MFP' in obs[8])
        self.assertTrue(str(run_prefix) in obs[10])
        self.assertTrue('AUTO' in obs[12])
        self.assertTrue('4' in obs[14])
        self.assertTrue('1723' in obs[16])
        self.assertTrue(str('-safe') in obs[17])
        self.assertTrue(str('-fast') in obs[18])
        self.assertTrue(str('1000') in obs[20])
        self.assertTrue(str('-abayes') in obs[21])
        self.assertTrue(str('1000') in obs[23])
        self.assertTrue(str('-allnni') in obs[24])
        self.assertTrue(str('200') in obs[26])
        self.assertTrue(str('30') in obs[28])
        self.assertTrue(str('10') in obs[30])
        self.assertTrue(str('80') in obs[32])
        self.assertTrue(str('300') in obs[34])
        self.assertTrue(str('0.55') in obs[36])
        self.assertTrue(str('8') in obs[38])

    def test_build_iqtree_ufbs_command(self):
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with tempfile.TemporaryDirectory() as temp_dir:
            run_prefix = os.path.join(temp_dir, 'q2iqtreeufboot')
            with redirected_stdio(stderr=os.devnull):
                obs = _build_iqtree_ufbs_command(input_sequences,
                                                 seed=1723,
                                                 n_cores=0,
                                                 n_cores_max=6,
                                                 n_runs=5,
                                                 bootstrap_replicates=2000,
                                                 substitution_model='MFP',
                                                 run_prefix=run_prefix,
                                                 dtype='DNA',
                                                 safe='True',
                                                 allnni='True',
                                                 alrt=500,
                                                 abayes=True,
                                                 lbp=400,
                                                 bnni=True,
                                                 n_init_pars_trees=200,
                                                 n_top_init_trees=30,
                                                 n_best_retain_trees=10,
                                                 stop_iter=300,
                                                 perturb_nni_strength=0.55,
                                                 spr_radius=8,
                                                 n_max_ufboot_iter=600,
                                                 n_ufboot_steps=80,
                                                 min_cor_ufboot=0.66,
                                                 ep_break_ufboot=0.51)
        self.assertTrue('2000' in obs[2])
        self.assertTrue('DNA' in obs[4])
        self.assertTrue('5' in obs[6])
        self.assertTrue(str(input_sequences) in str(obs[8]))
        self.assertTrue('MFP' in obs[10])
        self.assertTrue(str(run_prefix) in obs[12])
        self.assertTrue('AUTO' in obs[14])
        self.assertTrue('6' in obs[16])
        self.assertTrue('1723' in obs[18])
        self.assertTrue('-safe' in obs[19])
        self.assertTrue('-allnni' in obs[20])
        self.assertTrue('500' in obs[22])
        self.assertTrue('-abayes' in obs[23])
        self.assertTrue('400' in obs[25])
        self.assertTrue('-bnni' in obs[26])
        self.assertTrue('200' in obs[28])
        self.assertTrue(str('30') in obs[30])
        self.assertTrue(str('10') in obs[32])
        self.assertTrue(str('300') in obs[34])
        self.assertTrue(str('0.55') in obs[36])
        self.assertTrue(str('8') in obs[38])
        self.assertTrue(str('600') in obs[40])
        self.assertTrue(str('80') in obs[42])
        self.assertTrue(str('0.66') in obs[44])
        self.assertTrue(str('0.51') in obs[46])

    def test_iqtree_ultrafast_bootstrap(self):
        # Test that output tree is made.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = iqtree_ultrafast_bootstrap(input_sequences)
        obs_tree = skbio.TreeNode.read(str(obs))
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_iqtree_ultrafast_bootstrap_singlebranch_methods(self):
        # Comparing branch support to manually constructed tree
        # using the following command:
        #  iqtree -s aligned-dna-sequences-3.fasta -alrt 1500 -lbp 1500
        #         -abayes -bb 1500 -m 'HKY' -seed 1723
        # Here I am simply checking if the support values are identical
        # to the manual run. Also check for number of values.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')

        with redirected_stdio(stderr=os.devnull):
            obs = iqtree_ultrafast_bootstrap(input_sequences, seed=1723,
                                             substitution_model='HKY',
                                             alrt=1500, lbp=1500,
                                             abayes=True,
                                             bootstrap_replicates=1500)
        obs_tree = skbio.TreeNode.read(str(obs), convert_underscores=False)
        obs_supp = [node.name for node in obs_tree.non_tips()]

        exp_tree = skbio.TreeNode.read(self.get_data_path('test5.tre'))
        exp_supp = [node.name for node in exp_tree.non_tips()]

        # iter through all support values for each node and check if they
        # are at least within 1 support unit away from each other
        zipped = zip_longest(obs_supp, exp_supp)
        for (branch_sup_obs, branch_sup_exp) in zipped:
            obs_sl = [float(n) for n in branch_sup_obs.split("/")]
            exp_sl = [float(n) for n in branch_sup_exp.split("/")]
            for (obs_f, exp_f) in zip_longest(obs_sl, exp_sl):
                if math.isclose(obs_f, exp_f, abs_tol=1):
                    continue
                else:
                    raise ValueError(
                        "Observed branch support values \'{}\' are not equal "
                        "to the expected values \'{}\'.".format(
                                            branch_sup_obs, branch_sup_exp))

    def test_iqtree_ultrafast_bootstrap_safe_allnni(self):
        # Test that output tree is made.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = iqtree_ultrafast_bootstrap(input_sequences, safe='True',
                                             allnni='True')
        obs_tree = skbio.TreeNode.read(str(obs))
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_iqtree_ultrafast_bootstrap_auto_threads(self):
        # Test that output tree is made after auto optimizing threads.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = iqtree_ultrafast_bootstrap(input_sequences, n_cores=0)
        obs_tree = skbio.TreeNode.read(str(obs))
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))

    def test_iqtree_ultrafast_bootstrap_auto_threads_max(self):
        # Test that output tree is made after auto optimizing threads.
        # Reads tree output and compares tip labels to expected labels.
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = iqtree_ultrafast_bootstrap(input_sequences, n_cores=0,
                                             n_cores_max=2)
        obs_tree = skbio.TreeNode.read(str(obs))
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515', 'GCA000454205',
                              'GCA000473545', 'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985', 'GCA900007555']))


class IqtreeRunCommandTests(TestPluginBase):

    package = 'q2_phylogeny.tests'

    def test_run_not_verbose(self):
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        aligned_fp = str(input_sequences)

        with tempfile.TemporaryDirectory() as temp_dir:
            run_prefix = os.path.join(temp_dir, 'q2iqtree')
            cmd = ['iqtree',
                   '-m', 'HKY',
                   '-seed', '1723',
                   '-s', aligned_fp,
                   '-pre', run_prefix,
                   '-nt', '2']

            with redirected_stdio(stderr=os.devnull):
                run_command(cmd, verbose=False)
            obs_tree_fp = run_prefix + '.treefile'
            obs_tree = skbio.TreeNode.read(str(obs_tree_fp),
                                           convert_underscores=False)
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515',
                              'GCA000454205', 'GCA000473545',
                              'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985',
                              'GCA900007555']))

    def test_run_ultrafast_bs_not_verbose(self):
        input_fp = self.get_data_path('aligned-dna-sequences-3.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        aligned_fp = str(input_sequences)

        with tempfile.TemporaryDirectory() as temp_dir:
            run_prefix = os.path.join(temp_dir, 'q2iqtreeufboot')
            cmd = ['iqtree',
                   '-m', 'HKY',
                   '-seed', '1723',
                   '-bb', '1000',
                   '-s', aligned_fp,
                   '-pre', run_prefix,
                   '-nt', '2']

            with redirected_stdio(stderr=os.devnull):
                run_command(cmd, verbose=False)
            obs_tree_fp = run_prefix + '.treefile'
            obs_tree = skbio.TreeNode.read(str(obs_tree_fp),
                                           convert_underscores=False)
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        self.assertEqual(set(tip_names),
                         set(['GCA001510755', 'GCA001045515',
                              'GCA000454205', 'GCA000473545',
                              'GCA000196255', 'GCA000686145',
                              'GCA001950115', 'GCA001971985',
                              'GCA900007555']))


if __name__ == "__main__":
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from ._util import midpoint_root, robinson_foulds
from ._fasttree import fasttree
from ._raxml import raxml, raxml_rapid_bootstrap
from ._iqtree import iqtree, iqtree_ultrafast_bootstrap
from ._filter import filter_table, filter_tree
from ._version import get_versions
from ._align_to_tree_mafft_fasttree import align_to_tree_mafft_fasttree
from ._align_to_tree_mafft_iqtree import align_to_tree_mafft_iqtree
from ._align_to_tree_mafft_raxml import align_to_tree_mafft_raxml

__version__ = get_versions()['version']
del get_versions

__all__ = ["midpoint_root", "fasttree", "align_to_tree_mafft_fasttree",
           "raxml", "raxml_rapid_bootstrap", "iqtree", "filter_table",
           "iqtree_ultrafast_bootstrap", "align_to_tree_mafft_iqtree",
           "align_to_tree_mafft_raxml", "robinson_foulds", 'filter_tree']



# ----------------------------------------------------------------------------
# Copyright (c) 2016-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import unittest
import skbio
import subprocess

from qiime2.plugin.testing import TestPluginBase
from qiime2.util import redirected_stdio
from q2_types.feature_data import AlignedDNAFASTAFormat
from q2_types.tree import NewickFormat

from q2_phylogeny import fasttree
from q2_phylogeny._fasttree import run_command


class FastTreeTests(TestPluginBase):

    package = 'q2_phylogeny.tests'

    def test_fasttree(self):
        input_fp = self.get_data_path('aligned-dna-sequences-1.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = fasttree(input_sequences)
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids (the branch lengths can vary with
        # different versions of FastTree)
        obs_tree = skbio.TreeNode.read(str(obs))
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        tip_names.sort()
        self.assertEqual(tip_names, ['seq1', 'seq2'])

    def test_fasttree_underscore_ids(self):
        input_fp = self.get_data_path('aligned-dna-sequences-2.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = fasttree(input_sequences)
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids (the branch lengths can vary with
        # different versions of FastTree)
        obs_tree = skbio.TreeNode.read(str(obs))
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        tip_names.sort()
        self.assertEqual(tip_names, ['_s_e_q_1_', '_s_e_q_2_'])

    def test_fasttree_n_tips(self):
        input_fp = self.get_data_path('aligned-dna-sequences-1.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        with redirected_stdio(stderr=os.devnull):
            obs = fasttree(input_sequences, n_threads='auto')
        # load the resulting tree and test that it has the right number of
        # tips and the right tip ids (the branch lengths can vary with
        # different versions of FastTree, and threading can produce
        # non-deterministic trees)
        obs_tree = skbio.TreeNode.read(str(obs))
        tips = list(obs_tree.tips())
        tip_names = [t.name for t in tips]
        tip_names.sort()
        self.assertEqual(tip_names, ['seq1', 'seq2'])


# We are using pytest vs. unittest for this test
# The capfd arg below is a built-in pytest fixture that allows for
# captured stderr from all subprocesses that directly write
# to operating system level output
def test_fasttree_num_threads(capfd):
    # Rather than providing an actual filepath, we just run a 'help' command
    # Ideally, we supply --help in this param, but not available in FastTree
    # So we are using -expert instead, so that FastTree doesn't fail
    # ######################
    # Output preview/example:
    # ######################
    # Detailed usage for FastTree 2.1.10 Double precision (No SSE3):
    # FastTree [-nt] [-n 100] [-quote] [-pseudo | -pseudo 1.0]
    #            [-boot 1000 | -nosupport]
    #            [-intree starting_trees_file | -intree1 starting_tree_file]
    #            [-quiet | -nopr]
    #            [-nni 10] [-spr 2] [-noml | -mllen | -mlnni 10]
    #            [-mlacc 2] [-cat 20 | -nocat] [-gamma]
    #            [-slow | -fastest] [-2nd | -no2nd] [-slownni] [-seed 1253]
    #            [-top | -notop] [-topm 1.0 [-close 0.75] [-refresh 0.8]]
    #            [-matrix Matrix | -nomatrix] [-nj | -bionj]
    #            [-lg] [-wag] [-nt] [-gtr] [-gtrrates ac ag at cg ct gt]
    #            [-gtrfreq A C G T]
    #            [ -constraints constraintAlignment
    #            [ -constraintWeight 100.0 ] ]
    #            [-log logfile]
    #          [ alignment_file ]
    #         [ -out output_newick_file | > newick_tree]
    fasttree('-expert', n_threads=1)
    captured = capfd.readouterr()
    assert 'OpenMP' not in captured.err

    fasttree('-expert', n_threads=20)
    captured = capfd.readouterr()
    assert 'OpenMP (20 threads)' in captured.err

    # This test case ensures that when a user enters 'auto', the n_threads
    # var will still be set to the max available on their machine, even if
    # the OMP_NUM_THREADS env var has been set on their machine
    os.environ['OMP_NUM_THREADS'] = '2560'
    fasttree('-expert', n_threads='auto')
    captured = capfd.readouterr()
    assert 'OpenMP (2560 threads)' not in captured.err


class RunCommandTests(TestPluginBase):

    package = 'q2_phylogeny.tests'

    def test_failed_run(self):
        input_fp = self.get_data_path('aligned-dna-sequences-1.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        result = NewickFormat()
        aligned_fp = str(input_sequences)
        tree_fp = str(result)

        cmd = ['FastTree', '-nt', '-not-a-real-parameter', aligned_fp]
        with self.assertRaises(subprocess.CalledProcessError):
            with redirected_stdio(stderr=os.devnull):
                run_command(cmd, tree_fp)

    def test_failed_run_not_verbose(self):
        input_fp = self.get_data_path('aligned-dna-sequences-1.fasta')
        input_sequences = AlignedDNAFASTAFormat(input_fp, mode='r')
        result = NewickFormat()
        aligned_fp = str(input_sequences)
        tree_fp = str(result)

        cmd = ['FastTree', '-nt', '-not-a-real-parameter', aligned_fp]
        with self.assertRaises(subprocess.CalledProcessError):
            with redirected_stdio(stderr=os.devnull):
                run_command(cmd, tree_fp, verbose=False)


if __name__ == "__main__":
    unittest.main()
